/* $Id: version.h.in,v 1.6 2005/01/29 20:54:20 rocky Exp $ */
/** \file version.h 
 *
 *  \brief A file containing the libcdio package version
 *  number (76) and OS build name. 
 */

/*! CDIO_VERSION can as a string in programs to show what version is used. */
#define CDIO_VERSION "0.76 i486-slackware-linux-gnu"

/*! LIBCDIO_VERSION_NUM  can be used for testing in the C preprocessor */
#define LIBCDIO_VERSION_NUM 76
